# Estory --- Read from and write to event stores
# Copyright © 2021, 2022 Bioneland
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
import os
import re
import sys
from dataclasses import asdict
from pathlib import Path

import pkg_resources
import typer
from bl_event_sourcing.core import Event
from bl_event_sourcing_sqlalchemy.core import EVENT_REGISTRY, EventStore
from sqlalchemy import create_engine  # type: ignore
from sqlalchemy.orm import Session, sessionmaker  # type: ignore

__version__ = pkg_resources.get_distribution(__name__).version
app = typer.Typer()


class CannotGuessDsn(Exception):
    pass


@app.command()
def guess(dsn: str) -> None:
    typer.echo(guess_real_dsn(dsn))


@app.command()
def init(dsn: str) -> None:
    try:
        real_dsn = guess_real_dsn(dsn)
    except CannotGuessDsn:
        path = Path(dsn)
        path.touch()
        real_dsn = f"sqlite:///{dsn}"

    EVENT_REGISTRY.metadata.create_all(create_engine(real_dsn))


@app.command()
def read(dsn: str) -> None:
    store = build_store(build_session(guess_real_dsn(dsn)))
    for event in store.get_stream():
        e = asdict(event)
        typer.echo(json.dumps(e))


@app.command()
def write(dsn: str) -> None:
    session = build_session(guess_real_dsn(dsn))
    store = build_store(session)
    for line in sys.stdin:
        data = json.loads(line.strip())
        event = Event(**data)
        store.record(event)
        # FIXME speed up insertion!
        # https://docs.sqlalchemy.org/en/14/faq/performance.html#i-m-inserting-400-000-rows-with-the-orm-and-it-s-really-slow
        session.commit()


def guess_real_dsn(dsn: str) -> str:
    if real_dsn := guess_real_dsn_from_env(dsn):
        return real_dsn
    if real_dsn := guess_real_dsn_from_uri(dsn):
        return real_dsn
    if real_dsn := guess_real_dsn_from_file(dsn):
        return real_dsn
    raise CannotGuessDsn(dsn)


def guess_real_dsn_from_env(dsn: str) -> str:
    if uri := os.environ.get(f"ESTORY_DSN_{dsn}"):
        return uri
    return ""


def guess_real_dsn_from_uri(dsn: str) -> str:
    if re.search(r"^[^:]+://.+$", dsn):
        return dsn
    return ""


def guess_real_dsn_from_file(dsn: str) -> str:
    if Path(dsn).exists():
        return f"sqlite:///{dsn}"
    return ""


def build_session(dsn: str) -> Session:
    with sessionmaker(bind=create_engine(dsn))() as session:
        with session.begin():
            return session


def build_store(session: Session) -> EventStore:
    return EventStore(session)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['estory']

package_data = \
{'': ['*']}

install_requires = \
['bl-event-sourcing-sqlalchemy>=0.1.0',
 'bl-event-sourcing>=0.1.0',
 'typer>=0.4.1,<0.5.0']

extras_require = \
{'mysql': ['PyMySQL>=1.0.2,<2.0.0']}

entry_points = \
{'console_scripts': ['estory = estory:app']}

setup_kwargs = {
    'name': 'estory',
    'version': '0.1.2',
    'description': 'Read from and write to event stores',
    'long_description': '# estory — read from and write to event stores\n\n```console\n$ estory init <IDENTIFIER>\n$ estory read <IDENTIFIER>\n$ estory write <IDENTIFIER>\n$ estory guess <IDENTIFIER>\n```\n\n**IDENTIFIER** is either:\n\n- a proper DSN (`driver://user:password@host:port/database`)\xa0;\n- a file that will be interpreted as a DSN (`sqlite:///<FILE>`)\xa0;\n- a name matching an environment variable (`ESTORY_DSN_<NAME>`).\n\nTo initialise an event store:\n\n```console\n$ export ESTORY_DSN_TMP="sqlite:////tmp/event_store.sqlite"\n$ estory init TMP\n```\n\nTo write an event to a store:\n\n```console\n$ jo stream="uuid" version=1 name="EventName" data=$(jo id=1) unixtime=$(date +%s) | estory write TMP\n```\n\nTo read events from a store:\n\n```console\n$ estory read TMP\n{"stream": "uuid", "name": "EventName", "data": {"id": 1}, "unixtime": 1647876362, "who": "", "id": 1}\n```\n\nTo copy events from one store to another:\n\n```console\n$ export ESTORY_DSN_TMP2="sqlite:////tmp/another_event_store.sqlite"\n$ estory init TMP2\n$ estory read TMP | estory write TMP2\n```\n\nTo copy only some events from one store to another:\n\n```console\n$ estory read TMP | jq \'.|select(.id==2)\' | estory write TMP2\n```\n\nTo modify events while copying them from one store to another:\n\n```console\n$ estory read TMP | jq \'.|.name="EventNewName"\' | estory write TMP2\n```\n',
    'author': 'Tanguy Le Carrour',
    'author_email': 'tanguy@bioneland.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://git.easter-eggs.org/bioneland/estory',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
